from copy import deepcopy

# NumPy
import numpy as np
# SciPy
from scipy import interpolate, signal
# matplotlib
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from matplotlib.collections import PatchCollection
# AstroPy
import astropy.units as u
from astropy.time import Time
from astropy.table import QTable
from astropy.nddata import StdDevUncertainty
# specutils
from specutils import Spectrum1D
# drpy
from drpy import conf
from drpy.modeling import Poly1D, Spline1D
from drpy.plotting import plotFitting, _plotSpectrum1D
from drpy.validate import (_validateBool, _validateString, _validateInteger, 
                           _validateSpectrum, _validatePath)

from .center import _center1D_Gaussian, _refinePeakBases, _refinePeaks
from .io import loadStandardSpectrum, loadExtinctionCurve#, loadSpectrum1D

__all__ = ['dispcor', 'sensfunc', 'calibrate1d']


def dispcor(spectrum1d, reverse, reference, n_sub=20, refit=True, prominence=1e-3, 
            degree=1, maxiters=5, sigma_lower=None, sigma_upper=None, grow=False, 
            use_mask=False, title='dispcor', show=conf.fig_show, save=conf.fig_save, 
            path=conf.fig_path, **kwargs):
    """Dispersion correction.

    Parameters
    ----------
    spectrum1d : `~specutils.Spectrum1D` or `~numpy.ndarray`
        Input spectrum.

    reverse : bool
        Reverse the input spectrum or not. Set to `True` if the spectral axis of the 
        input spectrum is in reverse order.

    reference : `~specutils.Spectrum1D` or str
        File name of the reference spectrum.

    n_sub : int
        A pixel is divided into ``n_sub`` subpixels before cross-correlation.
        Default is `20`.

    refit : bool, optional
        Refit the dispersion solution or not. If `True`, lines in the input spectrum 
        are recentered and then used to fit the dispersion solution.

    degree : int, optional
        Degree of the fitting polynomial.
        Default is `3`.

    maxiters : int, optional
        Maximum number of sigma-clipping iterations to perform. If convergence is 
        achieved prior to ``maxiters`` iterations, the clipping iterations will stop. 
        Must be >= `0`. 
        Default is `5`.

    sigma_lower : scalar or `None`, optional
        Number of standard deviations to use as the lower bound for the clipping limit. 
        If `None` (default), `3` is used.

    sigma_upper : scalar or `None`, optional
        Number of standard deviations to use as the upper bound for the clipping limit. 
        If `None` (default), `3` is used.

    grow : float or `False`, optional
        Radius within which to mask the neighbouring pixels of those that fall outwith 
        the clipping limits (only applied along axis, if specified).

    use_mask : bool, optional
        If `True` and a mask array is attributed to ``spectrum1d``, the masked pixels are 
        interpolated before cross-correlation. 
        Default is `False`.

    Returns
    -------
    new_spectrum1d : `~specutils.Spectrum1D`
        Dispersion corrected spectrum.
    """
    
    
    new_spectrum1d, _, flux, _, mask = _validateSpectrum(
        spectrum1d, 'spectrum1d', False, use_mask)

    # ``flux`` and ``mask`` may be 2-dimensional arrays, though each one may only have 
    # one row. Use flatten to get rid of the additional dimension.
    if flux.ndim > 1:
        flux = flux.flatten()[:flux.shape[-1]]
        mask = mask.flatten()[:mask.shape[-1]]

    index = np.arange(flux.shape[0])

    # Interpolate bad pixels
    flux = interpolate.interp1d(
        index[~mask], flux[~mask], bounds_error=False, fill_value='extrapolate', 
        assume_sorted=True)(index)

    flux /= np.nanmax(flux)
    
    _validateBool(reverse, 'reverse')

    if reverse:
        flux = flux[::-1]

    # Parse reference spectrum
    if isinstance(reference, Spectrum1D):
        spectrum1d_ref = deepcopy(reference)
    else:
        spectrum1d_ref = Spectrum1D.read(reference)
    spectral_axis_ref = spectrum1d_ref.spectral_axis.value
    unit_spectral_axis = spectrum1d_ref.spectral_axis.unit
    flux_ref = spectrum1d_ref.flux.value
    flux_ref /= np.nanmax(flux_ref)

    n_pix = spectral_axis_ref.shape[0]

    index_ref = np.arange(n_pix)

    n_sub = _validateInteger(n_sub, 'n_sub', (1, None), (True, None))

    # Of type `~numpy.float64`, although still named as an index array.
    index_dense = np.linspace(0, index[-1], index[-1] * n_sub + 1)
    flux_dense = interpolate.interp1d(index, flux, assume_sorted=True)(index_dense)

    # Of type `~numpy.float64`, although still named as an index array.
    index_ref_dense = np.linspace(0, index_ref[-1], index_ref[-1] * n_sub + 1)
    flux_ref_dense = interpolate.interp1d(
        index_ref, flux_ref, assume_sorted=True)(index_ref_dense)

    # Here assumes that the zeropoint shift is no more than ±(n/4)
    n_ext = index_ref[-1] // 4

    # Reference spectrum
    index_ref_dense_ext = np.linspace(
        0, index_ref[-1] + 2 * n_ext, (index_ref[-1] + 2 * n_ext) * n_sub + 1) - n_ext
    flux_ref_dense_ext = np.hstack([
        flux_ref_dense[-n_ext * n_sub:], flux_ref_dense, flux_ref_dense[:n_ext * n_sub]
    ])

    # Get zeropoint shift through cross-correlation
    index_max = signal.correlate(
        flux_ref_dense_ext, flux_dense, mode='valid', method='auto').argmax()
    # The shift value is just the index of the maximum in the ``index_ref_dense_ext``
    delta = index_ref_dense_ext[index_max]

    _validateBool(refit, 'refit')

    if refit:

        # Find peaks in the reference spectrum
        peaks, properties = signal.find_peaks(
            x=flux_ref, prominence=prominence, **kwargs)
        # Refine peaks in the reference spectrum
        heights = flux_ref[peaks]
        refined_left_bases, refined_right_bases = _refinePeakBases(
            peaks=peaks, left_bases=properties['left_bases'], 
            right_bases=properties['right_bases'], n_peak=peaks.shape[0], copy=True)
        refined_peaks_ref, refined_index_ref = _refinePeaks(
            flux_ref, peaks, heights, refined_left_bases, refined_right_bases, 1)
    
        # Reidentify in the input spectrum
        shifted_peaks = refined_peaks_ref - delta
        heights = heights[refined_index_ref]
        shifted_left_bases = refined_left_bases[refined_index_ref] - delta
        shifted_right_bases = refined_right_bases[refined_index_ref] - delta
        refined_peaks, refined_index = _refinePeaks(
            flux, shifted_peaks, heights, shifted_left_bases, shifted_right_bases, 1)

        # Fit polynomial
        delta_peaks = refined_peaks_ref[refined_index] - refined_peaks
        spl, residual, threshold_lower, threshold_upper, master_mask = Poly1D(
            x=refined_peaks, y=delta_peaks, deg=degree, maxiters=maxiters, 
            sigma_lower=sigma_lower, sigma_upper=sigma_upper, grow=grow, 
            use_relative=False)
        delta = spl(index)

        rms = np.sqrt((residual[~master_mask]**2).sum() / (~master_mask).sum())

        # print(f'dispersion solution rms = {rms:.3}')

        plotFitting(
            x=refined_peaks, y=delta_peaks, m=master_mask, x_fit=index, y_fit=delta, 
            r=residual, threshold_lower=threshold_lower, 
            threshold_upper=threshold_upper, xlabel='dispersion axis [px]', 
            ylabel='difference [px]', title='dispcor', show=show, save=save, 
            path=path, use_relative=False)

    # !!! The extrapolated wavelengths are not reliable !!!
    spectral_axis = interpolate.interp1d(
        index_ref, spectral_axis_ref, bounds_error=False, fill_value='extrapolate', 
        assume_sorted=True)(index + delta)

    # Plot
    _validateBool(show, 'show')
    _validateBool(save, 'save')

    if show | save:

        _validateString(title, 'title')
        if title != 'dispcor':
            title = f'{title} dispcor'

        # Peak detection plot

        # Split into subplots
        n_subplot = 2

        fig, ax = plt.subplots(n_subplot, 1, dpi=100)

        length = n_pix // n_subplot + 1
        for i in range(n_subplot):

            idx_start, idx_end = i * length, (i + 1) * length
            idx_peak = np.where(
                (idx_start <= refined_peaks_ref[refined_index]) 
                & (refined_peaks_ref[refined_index] < idx_end))[0]

            ax[i].step(
                index_ref[idx_start:idx_end], flux_ref[idx_start:idx_end], color='k', 
                ls='-', where='mid')
            for idx in idx_peak:
                ymin = heights[refined_index][idx] * 1.2
                ymax = heights[refined_index][idx] * 1.5
                ax[i].plot(
                    [refined_peaks_ref[refined_index][idx], 
                     refined_peaks_ref[refined_index][idx]], [ymin, ymax], 'r-', 
                    lw=1.5)

            # Settings
            ax[i].grid(axis='both', color='0.95', zorder=-1)
            ax[i].set_xlim(idx_start, idx_end)
            ax[i].set_yscale('log')
            ax[i].tick_params(
                which='major', direction='in', top=True, right=True, length=5, 
                width=1.5, labelsize=12)
            ax[i].tick_params(
                which='minor', direction='in', top=True, right=True, length=3, 
                width=1.5, labelsize=12)
            ax[i].set_ylabel('normalized intensity', fontsize=16)
        ax[-1].set_xlabel('dispersion axis [px]', fontsize=16)
        ax[0].set_title(f'{title} peak detection', fontsize=16)
        fig.align_ylabels()
        fig.set_figheight(fig.get_figwidth() * n_subplot / 2)
        fig.tight_layout()

        # Save
        if save:
            fig_path = _validatePath(path, f'{title} peak detection')
            plt.savefig(fig_path, dpi=100)
        
        if show:
            plt.show()

        plt.close()
        
        # Comparison
        xlabel = f'spectral axis [{unit_spectral_axis.to_string()}]'

        fig, ax = plt.subplots(1, 1, dpi=100)

        # Spectrum
        ax.step(
            spectral_axis, flux + 1, where='mid', color='C0', zorder=3, label='custom')
        ax.step(
            spectral_axis_ref, flux_ref, where='mid', color='C1', zorder=3, 
            label='reference')

        # Settings
        ax.grid(axis='both', color='0.95', zorder=-1)
        ax.set_xlim(spectral_axis[0], spectral_axis[-1])
        ax.tick_params(
            which='major', direction='in', top=True, right=True, length=5, width=1.5, 
            labelsize=12)
        ax.set_xlabel(xlabel, fontsize=16)
        ax.set_ylabel('flux', fontsize=16)
        ax.legend(fontsize=16)
        ax.set_title(title, fontsize=16)
        fig.set_figheight(fig.get_figwidth() * 2 / 3)
        fig.tight_layout()
        
        if save:
            fig_path = _validatePath(path, title)
            plt.savefig(fig_path, dpi=100)

        if show:
            plt.show()

        plt.close()

    if reverse:
        spectral_axis = spectral_axis[::-1]

    spectral_axis *= unit_spectral_axis

    if 'header' in new_spectrum1d.meta:
        meta = new_spectrum1d.meta.copy()

    else:
        meta = {'header': dict()}

    meta['header']['DISPCOR'] = ('{} Dispersion correction.'.format(
        Time.now().to_value('iso', subfmt='date_hm')))

    # Output
    # `__setattr__` of `~specutils.Spectrum1D` is disabled
    new_spectrum1d = Spectrum1D(
        spectral_axis=spectral_axis, flux=new_spectrum1d.flux, 
        uncertainty=new_spectrum1d.uncertainty, mask=new_spectrum1d.mask, meta=meta)

    return new_spectrum1d


# Old version
def _dispcor(spectrum1d, reverse, reference, n_piece=3, refit=True, maxiters=5, 
            sigma_lower=None, sigma_upper=None, grow=False, use_mask=False, 
            title='dispcor', show=conf.fig_show, save=conf.fig_save, 
            path=conf.fig_path):
    """Dispersion correction.

    Parameters
    ----------
    spectrum1d : `~specutils.Spectrum1D` or `~numpy.ndarray`
        Input spectrum.

    reverse : bool
        Reverse the input spectrum or not. Set to `True` if the spectral axis of the 
        input spectrum is in reverse order.

    reference : str
        File name of the reference spectrum.

    n_piece : int, optional
        Number of spline pieces. Lengths are all equal. Must be positive.
        Default is `3`.

    refit : bool, optional
        Refit the dispersion solution or not. If `True`, lines in the input spectrum 
        are recentered and then used to fit the dispersion solution.

    maxiters : int, optional
        Maximum number of sigma-clipping iterations to perform. If convergence is 
        achieved prior to ``maxiters`` iterations, the clipping iterations will stop. 
        Must be >= `0`. 
        Default is `5`.

    sigma_lower : scalar or `None`, optional
        Number of standard deviations to use as the lower bound for the clipping limit. 
        If `None` (default), `3` is used.

    sigma_upper : scalar or `None`, optional
        Number of standard deviations to use as the upper bound for the clipping limit. 
        If `None` (default), `3` is used.

    grow : float or `False`, optional
        Radius within which to mask the neighbouring pixels of those that fall outwith 
        the clipping limits (only applied along axis, if specified).

    use_mask : bool, optional
        If `True` and a mask array is attributed to ``spectrum1d``, the masked pixels are 
        interpolated before cross-correlation. 
        Default is `False`.

    Returns
    -------
    new_spectrum1d : `~specutils.Spectrum1D`
        Dispersion corrected spectrum.
    """
    
    
    new_spectrum1d, _, flux, _, mask = _validateSpectrum(
        spectrum1d, 'spectrum1d', False, use_mask)

    # ``flux`` and ``mask`` may be 2-dimensional arrays, though each one may only have 
    # one row. Use flatten to get rid of the additional dimension.
    if flux.ndim > 1:
        flux = flux.flatten()[:flux.shape[-1]]
        mask = mask.flatten()[:mask.shape[-1]]

    index = np.arange(flux.shape[0])

    # Interpolate bad pixels
    flux = interpolate.interp1d(
        index[~mask], flux[~mask], bounds_error=False, fill_value='extrapolate', 
        assume_sorted=True)(index)

    flux /= flux.max()
    
    _validateBool(reverse, 'reverse')

    if reverse:
        flux = flux[::-1]

    spectrum1d_ref = loadSpectrum1D(reference, ext='spec')
    spectral_axis_ref = spectrum1d_ref.spectral_axis.value
    unit_spectral_axis = spectrum1d_ref.spectral_axis.unit
    flux_ref = spectrum1d_ref.flux.value
    index_ref = np.arange(spectral_axis_ref.shape[0])

    # A precision of 0.05 px for cross-correlation (can be changed)
    n_sub = 20

    # Of type `~numpy.float64`, although still named as an index array.
    index_dense = np.linspace(0, index[-1], index[-1] * n_sub + 1)
    flux_dense = interpolate.interp1d(index, flux, assume_sorted=True)(index_dense)

    # Of type `~numpy.float64`, although still named as an index array.
    index_ref_dense = np.linspace(0, index_ref[-1], index_ref[-1] * n_sub + 1)
    flux_ref_dense = interpolate.interp1d(
        index_ref, flux_ref, assume_sorted=True)(index_ref_dense)

    # Here assumes that the zeropoint shift is no more than ±(n/4)
    n_ext = index_ref[-1] // 4

    # Reference spectrum
    index_ref_dense_ext = np.linspace(
        0, index_ref[-1] + 2 * n_ext, (index_ref[-1] + 2 * n_ext) * n_sub + 1) - n_ext
    flux_ref_dense_ext = np.hstack([
        flux_ref_dense[-n_ext * n_sub:], flux_ref_dense, flux_ref_dense[:n_ext * n_sub]
    ])

    # Get zeropoint shift through cross-correlation
    index_max = signal.correlate(
        flux_ref_dense_ext, flux_dense, mode='valid', method='auto').argmax()
    # The shift value is just the index of the maximum in the ``index_ref_dense_ext``
    shift = index_ref_dense_ext[index_max]

    _validateBool(refit, 'refit')

    if not refit:
        # !!! The extrapolated wavelengths are not reliable !!!
        spectral_axis = interpolate.interp1d(
            index_ref, spectral_axis_ref, bounds_error=False, fill_value='extrapolate', 
            assume_sorted=True)(index + shift)

    else:
        # Load peak table
        peak_tbl = QTable.read(reference, format='fits', hdu='peak')

        # Reidentify
        shifted_peaks = peak_tbl['peaks'].value - shift
        shifted_left_bases = peak_tbl['left_bases'].value - shift
        shifted_right_bases = peak_tbl['right_bases'].value - shift
        heights = peak_tbl['heights'].value
        
        refined_left_bases, refined_right_bases = _refinePeakBases(
            peaks=shifted_peaks, left_bases=shifted_left_bases, 
            right_bases=shifted_right_bases, n_peak=shifted_peaks.shape[0], copy=True)

        refined_peaks, refined_index = _refinePeaks(
            flux, shifted_peaks, heights, refined_left_bases, refined_right_bases, 1)

        # Fit cubic spline function
        spectral_axis_peaks = peak_tbl['spectral_axis'].value[refined_index]
        spl, residual, threshold_lower, threshold_upper, master_mask = Spline1D(
            x=refined_peaks, y=spectral_axis_peaks, order=3, n_piece=n_piece, 
            maxiters=maxiters, sigma_lower=sigma_lower, sigma_upper=sigma_upper, grow=grow, 
            use_relative=False)
        spectral_axis = spl(index)

        rms = np.sqrt((residual[~master_mask]**2).sum() / (~master_mask).sum())

        print(f'dispersion solution rms = {rms:.3}')

        plotFitting(
            x=spectral_axis_peaks, y=refined_peaks, m=master_mask, x_fit=spectral_axis, 
            y_fit=index, r=residual, threshold_lower=threshold_lower, 
            threshold_upper=threshold_upper, 
            xlabel=f'spectral axis [{unit_spectral_axis.to_string()}]', 
            ylabel='dispersion axis [px]', title='dispersion solution', show=show, 
            save=save, path=path, use_relative=False)

    # Plot
    _validateBool(show, 'show')
    _validateBool(save, 'save')

    if show | save:

        _validateString(title, 'title')
        if title != 'dispcor':
            title = f'{title} dispcor'

        xlabel = f'spectral axis [{unit_spectral_axis.to_string()}]'

        fig, ax = plt.subplots(1, 1, dpi=100)

        # Spectrum
        ax.step(
            spectral_axis, flux + 1, where='mid', color='C0', zorder=3, label='custom')
        ax.step(
            spectral_axis_ref, flux_ref, where='mid', color='C1', zorder=3, 
            label='reference')

        # Settings
        ax.grid(axis='both', color='0.95', zorder=-1)
        ax.set_xlim(spectral_axis[0], spectral_axis[-1])
        ax.tick_params(
            which='major', direction='in', top=True, right=True, length=5, width=1.5, 
            labelsize=12)
        ax.set_xlabel(xlabel, fontsize=16)
        ax.set_ylabel('flux', fontsize=16)
        ax.legend(fontsize=16)
        ax.set_title(title, fontsize=16)
        fig.set_figheight(0.5 * fig.get_figwidth())
        fig.tight_layout()
        
        if save:
            fig_path = _validatePath(path, title)
            plt.savefig(fig_path, dpi=100)

        if show:
            plt.show()

        plt.close()

    if reverse:
        spectral_axis = spectral_axis[::-1]

    spectral_axis *= unit_spectral_axis

    if 'header' in new_spectrum1d.meta:
        meta = new_spectrum1d.meta.copy()

    else:
        meta = {'header': dict()}

    meta['header']['DISPCOR'] = ('{} Dispersion correction.'.format(
        Time.now().to_value('iso', subfmt='date_hm')))

    # Output
    # `__setattr__` of `~specutils.Spectrum1D` is disabled
    new_spectrum1d = Spectrum1D(
        spectral_axis=spectral_axis, flux=new_spectrum1d.flux, 
        uncertainty=new_spectrum1d.uncertainty, mask=new_spectrum1d.mask, meta=meta)

    return new_spectrum1d


def sensfunc(spectrum1d, exptime, airmass, extinct, standard, bandwid=None, 
             bandsep=None, n_piece=3, maxiters=5, sigma_lower=None, sigma_upper=None, 
             grow=False, use_mask=False, title='sensfunc', show=conf.fig_show, 
             save=conf.fig_save, path=conf.fig_path):
    """Create sensitivity function.

    Parameters
    ----------
    spectrum1d : `~specutils.Spectrum1D` or `~numpy.ndarray`
        Input spectrum.

    exptime : str or scalar
        Exposure time. Should be either a keyword in the header (str) or exposure time 
        itself (scalar).

    airmass : str or scalar
        Airmass. Should be either a keyword in the header (str) or airmass itself 
        (scalar).

    extinct : str, `~specutils.Spectrum1D` or `None`
        Extinction curve. Should be either a file name (str) or the extinction curve 
        itself (`~specutils.Spectrum1D`) if not `None`.

    standard : str or `~specutils.Spectrum1D`
        Standard spectrum. Should be either a file name (str) in the libaray or the 
        standard spectrum itself (`~specutils.Spectrum1D`).

    bandwid, bandsep : scalar or `None`, optional
        Bandpass widths and separations in wavelength units.

    n_piece : int, optional
        Number of spline pieces. Lengths are all equal. Must be positive.
        Default is `3`.

    maxiters : int, optional
        Maximum number of sigma-clipping iterations to perform. If convergence is 
        achieved prior to ``maxiters`` iterations, the clipping iterations will stop. 
        Must be >= `0`. 
        Default is `5`.

    sigma_lower : scalar or `None`, optional
        Number of standard deviations to use as the lower bound for the clipping limit. 
        If `None` (default), `3` is used.

    sigma_upper : scalar or `None`, optional
        Number of standard deviations to use as the upper bound for the clipping limit. 
        If `None` (default), `3` is used.

    grow : float or `False`, optional
        Radius within which to mask the neighbouring pixels of those that fall outwith 
        the clipping limits (only applied along axis, if specified).

    use_mask : bool, optional
        If `True` and a mask array is attributed to ``spectrum1d``, the masked pixels 
        are ignored in the fitting. 
        Default is `False`.

    Returns
    -------
    sens1d : `~specutils.Spectrum1D` or `~numpy.ndarray`
        Sensitivity function.
    """

    new_spectrum1d, _, flux_obs, uncertainty_obs, mask_obs = _validateSpectrum(
        spectrum1d, 'spectrum1d', False, use_mask)

    if isinstance(exptime, str):
        exptime = spectrum1d.meta['header'][exptime]

    # ``flux`` and ``mask`` may be 2-dimensional arrays, though each one may only have 
    # one row. Use flatten to get rid of additional dimensions.
    if flux_obs.ndim > 1:
        flux_obs = flux_obs.flatten()[:flux_obs.shape[-1]]
        mask_obs = mask_obs.flatten()[:mask_obs.shape[-1]]

    flux_obs /= exptime

    wavelength_obs = new_spectrum1d.wavelength.value.copy()
    bin_edges_obs = new_spectrum1d.bin_edges.to(u.AA).value.copy()
    bin_width_obs = np.diff(bin_edges_obs)

    # Check whether is reversed
    if wavelength_obs[0] > wavelength_obs[-1]:
        wavelength_obs = wavelength_obs[::-1] # [Angstrom]
        bin_edges_obs = bin_edges_obs[::-1]   # [Angstrom]
        bin_width_obs = -bin_width_obs[::-1]  # [Angstrom]
        flux_obs = flux_obs[::-1]             # [counts/s]
        mask_obs = mask_obs[::-1]
        isReversed = True
    else:
        isReversed = False

    n_obs = wavelength_obs.shape[0]

    idx_obs = np.arange(n_obs)

    # Library spectrum
    if isinstance(standard, str):
        spectrum_lib, bandpass_lib = loadStandardSpectrum(standard)

    elif isinstance(standard, Spectrum1D):
        spectrum_lib = deepcopy(standard)
        bandpass_lib = None

    else:
        raise TypeError(
            '``standard`` should be either a file name of a standard spectrum or a '
            '`~specutils.Spectrum1D` object.')

    wavelength_lib = spectrum_lib.wavelength.value # [Angstrom]
    flux_lib = spectrum_lib.flux.value             # [erg/s/cm2/Angstrom]
    if bandpass_lib is None:                       # [Angstrom]
        bandpass_lib = np.diff(spectrum_lib.bin_edges).to(u.AA).value
    else:
        bandpass_lib = bandpass_lib.value

    # Default
    if (bandwid is None) | (bandsep is None):
        bandpass_arr = np.vstack(
            [wavelength_lib - bandpass_lib / 2, wavelength_lib + bandpass_lib / 2]).T
        mask = (
            (wavelength_obs[0] <= bandpass_arr[:, 0]) 
            & (bandpass_arr[:, 1] <= wavelength_obs[-1])
        )
        wavelength = wavelength_lib[mask]
        flux_bp_lib = flux_lib[mask]
        bandpass_arr = bandpass_arr[mask]

    # Custom
    else:
        coverage = wavelength_obs[-1] - wavelength_obs[0]
        n_band = coverage // (bandwid + bandsep)
        if coverage - n_band * (bandwid + bandsep) >= bandwid:
            n_band += 1
        wavelength = (
            wavelength_obs[0] + bandwid / 2 + np.arange(n_band) * (bandwid + bandsep)
        )
        bandpass_arr = np.vstack([wavelength - bandwid / 2, wavelength + bandwid / 2]).T
        flux_bp_lib = np.zeros(bandpass_arr.shape[0])

        for i, bandpass in enumerate(bandpass_arr):

            # Edge pixels
            edges_lib = interpolate.interp1d(
                x=wavelength_lib, y=flux_lib, bounds_error=False, fill_value=np.nan, 
                assume_sorted=True)(bandpass)

            # Integral over the bandpass
            mask_bp_lib = (
                (bandpass[0] < wavelength_lib) & (wavelength_lib < bandpass[1])
            )
            x = np.hstack([bandpass[0], wavelength_lib[mask_bp_lib], bandpass[1]])
            y = np.hstack([edges_lib[0], flux_lib[mask_bp_lib], edges_lib[1]])
            flux_bp_lib[i] = np.trapz(y=y, x=x) / bandwid

    flux_bp_obs = np.zeros(bandpass_arr.shape[0]) # [counts/s/Angstrom]
    mask_bp_obs = np.zeros(bandpass_arr.shape[0], dtype=bool)

    for i, bandpass in enumerate(bandpass_arr):

        # Internal pixels
        mask = (
            (bandpass[0] < bin_edges_obs[:-1]) & (bin_edges_obs[1:] < bandpass[1])
        )
        flux_bp_obs[i] = flux_obs[mask].sum()
        mask_bp_obs[i] = np.any(mask_obs[mask])

        # Edge pixels

        # ``idx_start`` labels the pixel where ``bandpass[0]`` is in
        idx_start = idx_obs[bin_edges_obs[:-1] <= bandpass[0]][-1]

        # ``idx_end`` labels the pixel where ``bandpass[1]`` is in
        idx_end = idx_obs[bin_edges_obs[1:] >= bandpass[1]][0]

        # ``bandpass[0]`` and ``bandpass[1]`` are in the same pixel
        if idx_start == idx_end:
            flux_bp_obs[i] += (
                flux_obs[idx_end] * (bandpass[1] - bandpass[0]) / bin_width_obs[idx_end]
            )
            mask_bp_obs[i] |= mask_obs[idx_end]

        # ``bandpass[0]`` and ``bandpass[1]`` are in different pixels
        else:
            flux_bp_obs[i] += (
                flux_obs[idx_start] * (bin_edges_obs[idx_start + 1] - bandpass[0]) 
                / bin_width_obs[idx_start]
            )
            flux_bp_obs[i] += (
                flux_obs[idx_end] * (bandpass[1] - bin_edges_obs[idx_end]) 
                / bin_width_obs[idx_end]
            )
            mask_bp_obs[i] |= (mask_obs[idx_start] | mask_obs[idx_end])
    
        flux_bp_obs[i] /= (bandpass[1] - bandpass[0])

    # Negative values lead to NaNs here
    sens = 2.5 * np.log10(flux_bp_obs / flux_bp_lib) # 2.5 x log10(counts / (erg / cm2))

    # Extinction curve
    if extinct is not None:

        if isinstance(airmass, str):
            airmass = spectrum1d.meta['header'][airmass]

        if isinstance(extinct, str):
            spectrum_ext = loadExtinctionCurve(extinct)

        elif isinstance(extinct, Spectrum1D):
            spectrum_ext = deepcopy(extinct)

        else:
            raise TypeError(
                '``extinct`` should be either a file name of a extinction curve or a '
                '`~specutils.Spectrum1D` object if not `None`.')

        wavelength_ext = spectrum_ext.wavelength.value
        extinction = spectrum_ext.flux.value

        # Has to be `quadratic` ???
        extinction = interpolate.interp1d(
            x=wavelength_ext, y=extinction, kind='quadratic', bounds_error=False, 
            fill_value='extrapolate', assume_sorted=True)(wavelength)

        sens += (airmass * extinction)

    mask_sens = np.isnan(sens) | mask_bp_obs

    # Fit cubic spline function
    spl, residual, threshold_lower, threshold_upper, master_mask = Spline1D(
        x=wavelength, y=sens, m=mask_sens, order=3, n_piece=n_piece, maxiters=maxiters, 
        sigma_lower=sigma_lower, sigma_upper=sigma_upper, grow=grow, use_relative=False)

    sens_fit = spl(wavelength_obs)
    # Use RMS as uncertainty
    uncertainty_sens_fit = np.full(n_obs, residual[~master_mask].std(ddof=1))
    
    out_of_range = (
        (wavelength_obs < wavelength[~master_mask][0]) | 
        (wavelength_obs > wavelength[~master_mask][-1])
    )
    sens_fit[out_of_range] = np.nan
    uncertainty_sens_fit[out_of_range] = np.nan

    # Plot
    _validateBool(show, 'show')
    _validateBool(save, 'save')

    if show | save:

        _validateString(title, 'title')

        fig, ax = plt.subplots(1, 1, dpi=100)

        # Spectrum
        _plotSpectrum1D(
            ax, wavelength_obs, (flux_obs / bin_width_obs), xlabel='wavelength [A]')

        # Bandpasses
        ymin, ymax = ax.get_ylim()
        height = (ymax - ymin) * 0.06
        patches = list()
        for i, bandpass in enumerate(bandpass_arr):
            width = bandpass[1] - bandpass[0]
            rectangle = Rectangle(
                (bandpass[0], (flux_bp_obs[i] - height / 2)), width, height)
            patches.append(rectangle)
        patch_collection = PatchCollection(
            patches, facecolor='None', edgecolor='r', lw=1.5, zorder=2.5)
        ax.add_collection(patch_collection)

        # Settings
        ax.set_title(title, fontsize=16)
        fig.set_figheight(fig.get_figwidth() * 2 / 3)
        fig.tight_layout()

        if save:
            fig_path = _validatePath(path, title)
            plt.savefig(fig_path, dpi=100)

        if show:
            plt.show()

        plt.close()

    # Sensitivity function fitting
    plotFitting(
        x=wavelength, y=sens, m=master_mask, x_fit=wavelength_obs, y_fit=sens_fit, 
        r=residual, threshold_lower=threshold_lower, threshold_upper=threshold_upper, 
        xlabel='wavelength [A]', ylabel='sensitivity function', title=title, show=show, 
        save=save, path=path, use_relative=False)

    unit_sens_fit = new_spectrum1d.flux.unit / u.s / u.AA / spectrum_lib.flux.unit

    if isReversed:
        sens_fit = sens_fit[::-1] * unit_sens_fit
        uncertainty_sens_fit = StdDevUncertainty(uncertainty_sens_fit[::-1])
    else:
        sens_fit *= unit_sens_fit
        sens_fit_uncertainty = StdDevUncertainty(uncertainty_sens_fit)

    if 'header' in new_spectrum1d.meta:
        meta = new_spectrum1d.meta.copy()
    else:
        meta = {'header': dict()}
    # Add headers here
    meta['header']['EXPTIME'] = exptime
    meta['header']['AIRMASS'] = airmass
    meta['header']['SENSFUNC'] = '{}'.format(
        Time.now().to_value('iso', subfmt='date_hm'))

    # Output
    sens1d = Spectrum1D(
        spectral_axis=new_spectrum1d.spectral_axis, flux=sens_fit, 
        uncertainty=uncertainty_sens_fit, meta=meta)

    return sens1d, spl


def calibrate1d(spectrum1d, exptime, airmass, extinct, sens1d, use_uncertainty=False):
    """Apply a flux calibration to 1-dimensional spectra.
    
    ``spectrum1d`` and ``sens1d`` are assumed to have the same spectral axis, including 
    values, units and order (reversed or not).

    Parameters
    ----------
    spectrum1d : `~specutils.Spectrum1D` or `~numpy.ndarray`
        Input spectrum.

    exptime : str or scalar
        Exposure time. Should be either a keyword in the header (str) or exposure time 
        itself (scalar).

    airmass : str or scalar
        Airmass. Should be either a keyword in the header (str) or airmass itself 
        (scalar).

    extinct : str, `~specutils.Spectrum1D` or `None`
        Extinction curve. Should be either a file name (str) or the extinction curve 
        itself (`~specutils.Spectrum1D`) if not `None`.

    sens1d : `~specutils.Spectrum1D` or `~numpy.ndarray`
        Sensitivity function.

    use_uncertainty : bool, optional
        If `True`, the uncertainty array attributed to ``sens1d`` is propagated.
        Default is `False`.

    Returns
    -------
    new_spectrum1d : `~specutils.Spectrum1D` or `~numpy.ndarray`
        Calibrated spectrum.
    """

    # ``flux_obs`` can be multi-dimensional array
    new_spectrum1d, _, _, _, _ = _validateSpectrum(
        spectrum1d, 'spectrum1d', False, False)

    wavelength_obs = new_spectrum1d.wavelength                     # [Angstrom]
    bin_width = np.abs(np.diff(new_spectrum1d.bin_edges.to(u.AA))) # [Angstrom]

    if isinstance(exptime, str):
        exptime = spectrum1d.meta['header'][exptime] * u.s

    flux_obs = new_spectrum1d.flux / (exptime * bin_width) # [counts/s/Angstrom]

    if new_spectrum1d.uncertainty is not None:
        uncertainty_obs = StdDevUncertainty(
            new_spectrum1d.uncertainty.array / (exptime.value * bin_width.value))
    else:
        uncertainty_obs = None

    new_spectrum1d = Spectrum1D(
        spectral_axis=wavelength_obs, flux=flux_obs, uncertainty=uncertainty_obs, 
        mask=new_spectrum1d.mask, meta=new_spectrum1d.meta)

    _validateBool(use_uncertainty, 'use_uncertainty')

    new_sens1d, _, sens, uncertainty_sens, _ = _validateSpectrum(
        sens1d, 'sens1d', use_uncertainty, False)

    wavelength_sens = new_sens1d.wavelength.value # [Angstrom]

    # ``wavelength_sens`` is dimensionless
    if not (wavelength_obs.value == wavelength_sens).all():
        raise ValueError(
            'The spectral axes of ``spectrum1d`` and ``sens1d`` should be the same.')

    # ``sens``, ``uncertainty_sens`` should be 1-dimensional arrays. 
    # Use flatten to get rid of additional dimensions.
    if sens.ndim > 1:
        sens = sens.flatten()[:sens.shape[-1]]
        uncertainty_sens = uncertainty_sens.flatten()[:uncertainty_sens.shape[-1]]

    # Extinction curve
    if extinct is not None:

        if isinstance(airmass, str):
            airmass = spectrum1d.meta['header'][airmass]

        if isinstance(extinct, str):
            spectrum_ext = loadExtinctionCurve(extinct)

        elif isinstance(extinct, Spectrum1D):
            spectrum_ext = deepcopy(extinct)

        else:
            raise TypeError(
                '``extinct`` should be either a file name of a extinction curve or a '
                '`~specutils.Spectrum1D` object if not `None`.')

        wavelength_ext = spectrum_ext.wavelength.value
        extinction = spectrum_ext.flux.value

        extinction = interpolate.interp1d(
            x=wavelength_ext, y=extinction, kind='quadratic', bounds_error=False, 
            fill_value='extrapolate', assume_sorted=True)(wavelength_sens)

        sens -= (airmass * extinction)

    sens = 10**(0.4 * sens) # [counts / (erg / cm2)]
    uncertainty_sens *= 0.4 * np.log(10) * sens # [counts / (erg / cm2)]

    new_sens1d = Spectrum1D(
        spectral_axis=(wavelength_sens * u.AA), flux=(sens * new_sens1d.flux.unit), 
        uncertainty=StdDevUncertainty(uncertainty_sens), meta=new_sens1d.meta)

    # Calibrate
    calibrated_spectrum1d = new_spectrum1d / new_sens1d
    
    # Output
    if new_spectrum1d.uncertainty is None:
        calibrated_spectrum1d.uncertainty = None

    if 'header' in new_spectrum1d.meta:
        meta = new_spectrum1d.meta.copy()
    else:
        meta = {'header': dict()}
    # Add headers here
    meta['header']['EXPTIME'] = exptime.value
    meta['header']['AIRMASS'] = airmass
    meta['header']['CALIBRAT'] = '{} Calibrated'.format(
        Time.now().to_value('iso', subfmt='date_hm'))

    calibrated_spectrum1d.meta = meta

    return calibrated_spectrum1d


import warnings

# NumPy
import numpy as np
# SciPy
from scipy.optimize import curve_fit, OptimizeWarning
# matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
# AstroPy
from astropy.nddata import Cutout2D
from astropy.stats import gaussian_sigma_to_fwhm, sigma_clip
# photutils
from photutils import CircularAperture, RectangularAperture
from photutils.background import Background2D
from photutils.centroids import centroid_com
from photutils.detection import find_peaks
# drpy
from drpy import conf
from drpy.modeling.function import CircularGaussian
from drpy.plotting import _plot2d
from drpy.validate import (_validateString, _validateBool, _validateRange, 
                           _validateInteger, _validateCCD, _validatePath)

# Set plot parameters
plt.rcParams['figure.figsize'] = [conf.fig_width, conf.fig_width]
plt.rcParams['axes.linewidth'] = 1.5
plt.rcParams['mathtext.fontset'] = 'stix'
plt.rcParams['font.family'] = 'STIXGeneral'

__all__ = ['getFWHM']


def getFWHM(ccd, box_size, n_sigma, aper_radius, saturation, n_peak, n_iter=5, 
            sigma_lower=None, sigma_upper=None, use_mask=False, title='fwhm', 
            show=conf.fig_show, save=conf.fig_save, path=conf.fig_path, **kwargs):
    """Estimate mean FWHM of the sources in the input image.

    The background of the input image is first modeled by a 2-dimensional background 
    estimator, and then subtracted from the input image. Detect sources in the 
    background subtracted image. Fit circular Gaussian profiles to the detected 
    sources.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData` or `~numpy.ndarray`
        Input image.

    box_size : int or array_like
        The box size along each axis. If ``box_size`` is a scalar then a square box of 
        size ``box_size`` will be used. If ``box_size`` has two elements, they must be 
        in (ny, nx) order. (for details see https://photutils.readthedocs.io/en/stable/
        api/photutils.background.Background2D.html)

    n_sigma : scalar
        The detection threshold is defined as ``n_sigma`` * background RMS.

    aper_radius : int
        Radius of a source in the input image. It is recommended to set comparable to 
        or slightly larger than the FWHM.

    saturation : scalar
        Saturation level of pixels.

    n_peak : int
        Number of peaks. When number of peaks exceeds ``n_peak``, the peaks with the 
        highest peak intensities are used in determining FWHM.

    n_iter : int, optional
        Number of sigma slipping iterations. Must be >= `0`.
        Default is `5`.

    sigma_lower : scalar or `None`, optional
        Number of standard deviations to use as the lower bound for the clipping limit. 
        If `None` (default), `3` is used.

    sigma_upper : scalar or `None`, optional
        Number of standard deviations to use as the upper bound for the clipping limit. 
        If `None` (default), `3` is used.

    use_mask : bool, optional
        If `True` and a mask array is attributed to ``ccd``, the masked pixels are 
        ignored in the background estimation and source detection. 
        Default is `False`.

    Returns
    -------
    fwhm : scalar
        Mean FWHM of the detected sources.

    fwhm_err : scalar
        Standard deviation of FWHM.
    """

    _validateInteger(aper_radius, 'aper_radius', (1, None), (True, None))

    # Assumes that ``aper_radius`` ~ FWHM ~ 2.3σ. Thus, for a circular Gaussian 
    # profile, almost 100% of the energy are concentrated in the aperture.
    aper_size = 2 * aper_radius + 1

    _validateBool(use_mask, 'use_mask')

    nccd, data_arr, _, mask_arr = _validateCCD(ccd, 'ccd', False, use_mask, False)

    # Background estimation
    bkg = Background2D(data=data_arr, box_size=box_size, mask=mask_arr, **kwargs)
    bkg_value = bkg.background
    bkgrms_value = bkg.background_rms

    _validateRange(n_sigma, 'n_sigma', (0, None), (False, None))

    # Source detection
    threshold = n_sigma * bkgrms_value
    data_arr_bkgsb = data_arr - bkg_value
    peak_table = find_peaks(
        data=data_arr_bkgsb, threshold=threshold, box_size=aper_size, footprint=None, 
        mask=mask_arr, border_width=(2 * aper_size - 1), centroid_func=centroid_com, 
        error=None, wcs=None)
    peak_table.sort('peak_value')
    peak_table.reverse()

    _validateRange(saturation, 'saturation', (0, None), (False, None))

    # Mask saturated sources
    isSaturated = peak_table['peak_value'].data >= saturation

    # Cutout size = 4 * ``aper_size`` - 3 = 8 * ``aper_radius`` + 1
    size = 4 * aper_size - 3

    # Mask sources too close to each other
    x, y = peak_table['x_centroid'].data, peak_table['y_centroid'].data
    isClose = np.zeros_like(x, dtype=bool)
    for i in range(x.shape[0]):
        if not isClose[i]:
            for j in range(i + 1, x.shape[0]):
                dist_sq = (x[j] - x[i])**2 + (y[j] - y[i])**2
                if dist_sq < size**2 / 2:
                    isClose[i], isClose[j] = True, True
    
    peak_table = peak_table[(~isSaturated) & (~isClose)]

    _validateInteger(n_peak, 'n_peak', (1, None), (True, None))

    # Remove redundant peaks
    peak_table = peak_table[:n_peak]

    # Update peak number (may become smaller)
    n_peak = len(peak_table)

    # Centroid
    x, y = peak_table['x_centroid'].data, peak_table['y_centroid'].data

    # X & Y grids
    Y, X = np.mgrid[-(size // 2):(size // 2 + 1), -(size // 2):(size // 2 + 1)]

    xy = (X.flatten(), Y.flatten())

    xy_min, xy_max = xy[0][0], xy[0][-1]

    position_arr = np.zeros((n_peak, 2)) + np.nan
    fwhm_arr = np.zeros(n_peak) + np.nan
    cutout2d_arr = np.zeros((n_peak, size, size))

    with warnings.catch_warnings():

        warnings.simplefilter('error', OptimizeWarning)

        for i in range(n_peak):

            # Cutout
            cutout2d = Cutout2D(
                data=data_arr_bkgsb, position=(x[i], y[i]), size=size, mode='strict')
            cutout2d_arr[i] = cutout2d.data / cutout2d.data.max()

            initial_guess = (1, 0, 0, aper_radius, 0)

            try:
                popt, _ = curve_fit(
                    CircularGaussian, xy, cutout2d_arr[i].flatten(), p0=initial_guess)

                a, x0, y0, sigma, b = popt

                if (a > 0) & (xy_min < x0 < xy_max) & (xy_min < y0 < xy_max):
                    position_arr[i] = x0 + size // 2, y0 + size // 2
                    fwhm_arr[i] = np.abs(sigma) * gaussian_sigma_to_fwhm
                    cutout2d_arr[i] -= b

                else:
                    raise RuntimeError('No feature fitted in the given region.')

            except (RuntimeError, TypeError, OptimizeWarning): # raise exception here
                pass

    # Output
    fwhm_arr_masked, threshold_lower, threshold_upper = sigma_clip(
        data=fwhm_arr, sigma_lower=sigma_lower, sigma_upper=sigma_upper, 
        maxiters=n_iter, stdfunc='mad_std', masked=True, return_bounds=True)
    fwhm, fwhm_err = fwhm_arr_masked.mean(), fwhm_arr_masked.std(ddof=1)

    if fwhm_arr_masked.mask.sum() > 0:
        warnings.warn(
            'Fitting for some sources failed (may caused by cosmic rays).', 
            RuntimeWarning
        )

    # Plot
    _validateBool(show, 'show')
    _validateBool(save, 'save')

    if show | save:

        _validateString(title, 'title')
        if title != 'fwhm':
            title = [f'{title} fwhm (source detection)', 
                     f'{title} fwhm (cutoff)']
        else:
            title = [f'fwhm (source detection)', f'fwhm (cutoff)']

        # Source detection plot
        fig, ax = plt.subplots(1, 1, dpi=100)

        # Image
        _plot2d(ax=ax, ccd=data_arr_bkgsb, cmap='Greys_r')
        (xmin, xmax), (ymin, ymax) = ax.get_xlim(), ax.get_ylim()
        # Mask
        ax.imshow(
            mask_arr, cmap='Greys', alpha=0.3 * mask_arr.astype(int), origin='lower')
        # Aperture
        apertures = RectangularAperture(
            positions=np.transpose([x, y]), w=size, h=size, theta=0)
        apertures[fwhm_arr_masked.mask].plot(ax, color='b', lw=1)
        apertures[~fwhm_arr_masked.mask].plot(ax, color='r', lw=1)

        # Setting
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        ax.annotate(
            'masked', xy=(0.75, 0.88), xycoords='axes fraction', fontsize=16, 
            color='b')
        ax.annotate(
            'detected', xy=(0.75, 0.93), xycoords='axes fraction', fontsize=16, 
            color='r')
        ax.set_title(title[0], fontsize=16)
        fig.set_figheight(
            (data_arr_bkgsb.shape[0] / data_arr_bkgsb.shape[1] * fig.get_figwidth()))
        fig.tight_layout()

        if save:
            fig_path = _validatePath(path, title[0])
            plt.savefig(fig_path, dpi=100)

        if show:
            plt.show()

        plt.close()

    # Cutout
    if save:

        fig_path = _validatePath(path, title[1], '.pdf')
        with PdfPages(fig_path, keep_empty=False) as pdf:

            for i in range(n_peak):

                if fwhm_arr_masked.mask[i]:
                    color = 'b'
                else:
                    color = 'r'

                fig, ax = plt.subplots(1, 1, figsize=(3.2, 3.2), dpi=100)

                # Image
                _plot2d(
                    ax=ax, ccd=cutout2d_arr[i], cmap='Greys_r', cbar=False, 
                    xlabel='column', ylabel='row')
                # Aperture
                apertures = CircularAperture(
                    (position_arr[i][0], position_arr[i][1]), r=(fwhm_arr[i] / 2))
                apertures.plot(ax, color=color, lw=1)
                # Centroid
                ax.plot(position_arr[i][0], position_arr[i][1], '+', color=color, ms=15)
                
                # Settings
                ax.tick_params(
                    which='major', direction='in', top=True, right=True, color='w', 
                    length=5, width=1, labelsize=12)
                ax.annotate(
                    'FWHM$=' + f'{fwhm_arr[i]:.2f}' + '\\,$px', xy=(0.05, 0.85), 
                    xycoords='axes fraction', color='w', fontsize=12)
                ax.set_title(f'cutoff at ({x[i]:.0f}, {y[i]:.0f})', fontsize=16)
                fig.tight_layout()

                pdf.savefig(fig, dpi=100)

                plt.close()
    
    return fwhm, fwhm_err


import os, textwrap, warnings
from copy import deepcopy

# NumPy
import numpy as np
# AstroPy
import astropy.units as u
from astropy.time import Time
from astropy.nddata import CCDData
# ccdproc
import ccdproc
# drpy
from drpy import conf
from drpy.utils import imstatistics
from drpy.validate import _validateBool

# Configurations
unit_ccddata = u.Unit(conf.unit_ccddata)
dtype_ccddata = np.dtype(conf.dtype_ccddata)

__all__ = ['CCDDataList']


class CCDDataList:
    """A class for batch processing."""

    @classmethod
    def read(cls, file_list, hdu=1, unit=unit_ccddata, dtype=dtype_ccddata):
        """Construct a `CCDDataList` object from a list of files.
        
        Parameters
        ----------
        file_list : array_like
            List of files.

        hdu : `int`, optional
            FITS extension from which a `CCDDataList` object should be initialized.
            Default is `1`.

        unit : `~astropy.units.Unit` or `str`, optional
            Unit of the image data.
            Default is `~astropy.units.adu`.

        dtype : `type` or `str`, optional
            Data type of the image data.
            Default is `~numpy.float32`.

        Returns
        -------
        ccddatalist : `CCDDataList`
            Constructed `CCDDataList` object.
        """

        ccddatalist = list()
        for file_name in file_list:

            # Load from file
            ccd = CCDData.read(file_name, hdu=hdu, unit=unit)

            # Convert to specified data type
            if ccd.data is not None:
                ccd.data = ccd.data.astype(dtype)

            if ccd.uncertainty is not None:
                ccd.uncertainty.arraye = ccd.uncertainty.array.astype(dtype)

            # Add or rewrite keyword `FILENAME` (used in `self.statistics`).
            ccd.header['FILENAME'] = os.path.split(file_name)[1]
            ccddatalist.append(ccd)

        ccddatalist = cls(ccddatalist)

        return ccddatalist


    def __init__(self, ccddatalist):
        """Construct a `CCDDataList` object.
        
        The size of each image should be the same. (This is not checked in the 
        current version.)

        Parameters
        ----------
        ccddatalist : Iterable
            List of `~astropy.nddata.CCDData` objects.
        """

        for ccd in ccddatalist:
            if not isinstance(ccd, CCDData):
                raise TypeError(
                    f'Invalid type `{type(ccd)}` for an item of `CCDDataList` object. '
                    '`~astropy.nddata.CCDData` is expected.')

        self._ccddatalist = deepcopy(ccddatalist)   


    def apply_over_ccd(self, func, *args, **kwargs):
        """Apply a function repeatedly over all the images.
        
        Parameters
        ----------
        func : function
            This function must take the to-be-processed image as its first argument.
        
        *args : tuple, optional
            Positional arguments passed through to ``func``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``func``.
        
        Returns
        -------
        ccddatalist : `CCDDataList`
            Processed images.
        """

        ccddatalist = list()

        for ccd in self._ccddatalist:

            ccddatalist.append(func(ccd, *args, **kwargs))

        ccddatalist = self.__class__(ccddatalist)

        return ccddatalist


    def transform_image(self, *args, **kwargs):
        """Transform images.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~ccdproc.transform_image``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.transform_image``.
        """

        return self.apply_over_ccd(ccdproc.transform_image, *args, **kwargs)

    
    def trim_image(self, *args, **kwargs):
        """Trim images to the dimensions indicated.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~ccdproc.trim_image``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.trim_image``.
        """

        return self.apply_over_ccd(ccdproc.trim_image, *args, **kwargs)


    def wcs_project(self, *args, **kwargs):
        """Project images onto a target WCS.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~ccdproc.wcs_project``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.wcs_project``.
        """

        return self.apply_over_ccd(ccdproc.wcs_project, *args, **kwargs)


    def subtract_bias(self, master, **kwargs):
        """Subtract master bias from images.
        
        Parameters
        ----------
        master : `~astropy.nddata.CCDData`
            Master image to be subtracted from input images.
        
        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.subtract_bias``.
        
        Returns
        -------
        ccddatalist : `CCDDataList`
            `CCDDataList` object with bias subtracted.
        """

        if 'add_keyword' not in kwargs:

            keywords_dict = {
                'ZEROCOR': '{} A level of {} is subtracted'.format(
                    Time.now().to_value('iso', subfmt='date_hm'), 
                    round(np.nanmean(master.data))
                )
            }

            ccddatalist =  self.apply_over_ccd(
                ccdproc.subtract_bias, master, add_keyword=keywords_dict, **kwargs)

        else:

            ccddatalist = self.apply_over_ccd(ccdproc.subtract_bias, master, **kwargs)
        
        return ccddatalist


    def subtract_dark(self, *args, **kwargs):
        """Subtract dark current from images.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~ccdproc.subtract_dark``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.subtract_dark``.
        """

        return self.apply_over_ccd(ccdproc.subtract_dark, *args, **kwargs)


    def subtract_overscan(self, *args, **kwargs):
        """Subtract the overscan region from images.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~ccdproc.subtract_overscan``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.subtract_overscan``.
        """

        return self.apply_over_ccd(ccdproc.subtract_overscan, *args, **kwargs)


    def flat_correct(self, flat, min_value=None, norm_value=None, **kwargs):
        """Correct images for flat fielding.
        
        Parameters
        ----------
        flat : `~astropy.nddata.CCDData`
            Flat field to apply to the data.

        min_value : float or `None`, optional
            Minimum value for flat field. The value can either be `None` and no minimum 
            value is applied to the flat or specified by a float which will replace all 
            values in the flat by the min_value.
            Default is `None`.

        norm_value : float or `None`, optional
            If not `None`, normalize flat field by this argument rather than the mean 
            of the image. This allows fixing several different flat fields to have the 
            same scale. If this value is negative or `0`, a `ValueError` is raised. 
            Default is `None`.
        
        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.flat_correct``.
        
        Returns
        -------
        ccddatalist : `CCDDataList`
            `CCDDataList` object with flat corrected.
        """

        if 'add_keyword' not in kwargs:

            keywords_dict = {
                'FLATCOR': '{} flat-fielded'.format(
                    Time.now().to_value('iso', subfmt='date_hm'))
            }

            ccddatalist = self.apply_over_ccd(
                ccdproc.flat_correct, flat, min_value, norm_value, 
                add_keyword=keywords_dict, **kwargs)

        else:

            ccddatalist = self.apply_over_ccd(
                ccdproc.flat_correct, flat, min_value, norm_value, **kwargs)

        return ccddatalist


    def create_deviation(self, *args, **kwargs):
        """Create uncertainty frames.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~ccdproc.create_deviation``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.create_deviation``.
        """

        return self.apply_over_ccd(ccdproc.create_deviation, *args, **kwargs)


    def gain_correct(self, *args, **kwargs):
        """Correct gain in images.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~ccdproc.gain_correct``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.gain_correct``.
        """

        return self.apply_over_ccd(ccdproc.gain_correct, *args, **kwargs)


    def ccd_process(self, *args, **kwargs):
        """Perform basic processing on ccd data.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~ccdproc.ccd_process``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.ccd_process``.
        """

        return self.apply_over_ccd(ccdproc.ccd_process, *args, **kwargs)


    def combine(self, *args, **kwargs):
        """Combine images.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~ccdproc.combine``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.combine``.
        
        Returns
        -------
        combined_ccd : `~astropy.nddata.CCDData`
            Combined image.
        """
        
        combined_ccd = ccdproc.combine(img_list=self._ccddatalist, *args, **kwargs)

        combined_ccd.header = self._ccddatalist[0].header
        combined_ccd.header['NCOMBINE'] = (self.__len__(), 'number of images combined')
        
        return combined_ccd


    def add(self, *args, **kwargs):
        """Perform addition.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~astropy.nddata.CCDData.add``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~astropy.nddata.CCDData.add``.
        """

        return self.apply_over_ccd(CCDData.add, *args, **kwargs)


    def subtract(self, *args, **kwargs):
        """Perform subtraction.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~astropy.nddata.CCDData.subtract``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~astropy.nddata.CCDData.subtract``.
        """

        return self.apply_over_ccd(CCDData.subtract, *args, **kwargs)


    def multiply(self, *args, **kwargs):
        """Perform multiplication.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~astropy.nddata.CCDData.multiply``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~astropy.nddata.CCDData.multiply``.
        """

        return self.apply_over_ccd(CCDData.multiply, *args, **kwargs)


    def divide(self, *args, **kwargs):
        """Perform division.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~astropy.nddata.CCDData.divide``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~astropy.nddata.CCDData.divide``.
        """

        return self.apply_over_ccd(CCDData.divide, *args, **kwargs)


    def cosmicray_lacosmic(self, use_mask=False, *args, **kwargs):
        """Remove cosmic rays.
        
        Parameters
        ----------
        use_mask : bool, optional
            If `True`, the attributed mask frames are used in identifying cosmic ray 
            pixels.
            Default is `False`.
        
        *args : tuple, optional
            Positional arguments passed through to ``~ccdproc.cosmicray_lacosmic``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.cosmicray_lacosmic``.
        
        Returns
        -------
        ccddatalist : `CCDDataList`
            `CCDDataList` object with cosmic rays removed.
        """

        _validateBool(use_mask, 'use_mask')

        keywords_dict = {
            'COSMICRA': '{} Laplacian Edge Detection'.format(
                Time.now().to_value('iso', subfmt='date_hm'))
        }

        ccddatalist = list()
        for ccd in self._ccddatalist:

            nccd = ccd.copy()

            if use_mask:
                if nccd.mask is None:
                    warnings.warn(
                        'The input mask is unavailable. All set unmasked.', 
                        RuntimeWarning)
                    nccd.mask = np.zeros_like(nccd.data, dtype=bool)
            else:
                nccd.mask = None

            # Of dtype float32
            ccd_cosmicray_removed = ccdproc.cosmicray_lacosmic(
                ccd=nccd, *args, **kwargs)

            for key, value in keywords_dict.items():
                ccd_cosmicray_removed.header[key] = value

            ccddatalist.append(ccd_cosmicray_removed)

        ccddatalist = self.__class__(ccddatalist)

        return ccddatalist


    def cosmicray_median(self, *args, **kwargs):
        """Remove cosmic rays.
        
        Parameters
        ----------
        *args : tuple, optional
            Positional arguments passed through to ``~ccdproc.cosmicray_median``.

        **kwargs : dict, optional
            Keyword arguments passed through to ``~ccdproc.cosmicray_median``.
        
        Returns
        -------
        ccddatalist : `CCDDataList`
            `CCDDataList` object with cosmic rays removed.
        """

        keywords_dict = {
            'COSMICRA': '{} median with sigma clipping'.format(
                Time.now().to_value('iso', subfmt='date_hm'))
        }

        ccddatalist = list()

        for ccd in self._ccddatalist:

            ccd_cosmicray_removed = ccdproc.cosmicray_median(ccd=ccd, *args, **kwargs)

            for key, value in keywords_dict.items():

                ccd_cosmicray_removed.header[key] = value

            ccddatalist.append(ccd_cosmicray_removed)

        ccddatalist = self.__class__(ccddatalist)

        return ccddatalist


    def statistics(self, verbose=False):
        """Statistics.

        Parameters
        ----------
        verbose : bool, optional
            If `True`, statistics table is printted before return.
            Default is `False`.

        Returns
        -------
        out : `~astropy.table.table.Table`
            Statistics table.
        """

        return imstatistics(self._ccddatalist, verbose=verbose)


    def to_list(self):
        """Convert to `list` object."""
        return self._ccddatalist


    def copy(self):
        """Return a copy of itself."""
        return self.__class__(self._ccddatalist)


    def __add__(self, operand2):
        """The same as `self.add`."""
        return self.add(operand2, handle_meta='first_found')


    def __sub__(self, operand2):
        """The same as `self.subtract`."""
        return self.subtract(operand2, handle_meta='first_found')


    def __mul__(self, operand2):
        """The same as `self.multiply`"""
        return self.multiply(operand2, handle_meta='first_found')


    def __truediv__(self, operand2):
        """The same as `self.divide`."""
        return self.divide(operand2, handle_meta='first_found')


    def __getitem__(self, item):

        if isinstance(item, (int, np.integer)):
            return self._ccddatalist[item]

        elif isinstance(item, slice):
            return self.__class__(self._ccddatalist[item])

        elif isinstance(item, (np.ndarray, list)):

            if isinstance(item, list):
                item = np.array(item)

            if item.ndim > 1:
                raise IndexError(
                    f'`CCDDataList` is 1-dimensional, but {item.ndim} were indexed.')

            else:
                if item.dtype == bool:

                    if item.size != self.__len__():
                        raise IndexError(
                            'Boolean index did not match the size of `CCDDataList`.')

                    else:
                        item = np.where(item)[0]

                ccddatalist = list()
                for i in item:
                    ccddatalist.append(self._ccddatalist[i])

                return self.__class__(ccddatalist)

        else:
            raise IndexError(
                f'Invalid type {type(item)} for `CCDDataList` item access.')


    def __repr__(self):

        if len(self._ccddatalist) == 0:
            core_string = ''

        elif len(self._ccddatalist) == 1:
            core_string = f'{self._ccddatalist[0]}'

        elif len(self._ccddatalist) == 2:
            core_string = f'{self._ccddatalist[0]},\n{self._ccddatalist[-1]}'

        elif len(self._ccddatalist) > 2:
            core_string = f'{self._ccddatalist[0]},\n...,\n{self._ccddatalist[-1]}'

        return (
            'CCDDataList([' + textwrap.indent(core_string, ' ' * 13)[13:] + '], length='
            f'{self.__len__()})'
        )


    def __len__(self):
        return len(self._ccddatalist)


    def __iter__(self):
        return iter(self._ccddatalist)


# Numpy
import numpy as np

__all__ = ['Gaussian1D', 'CircularGaussian']


def Gaussian1D(x, a, x0, sigma):
    """1D Gaussian function.

    Parameters
    ----------
    x : scalar or `~numpy.ndarray`
        Value or values to evaluate the Gaussian at.

    a : scalar
        Amplitude.

    x0 : scalar
        Gaussian center.

    sigma : scalar
        Standard deviation.

    Returns
    -------
    out: scalar or `~numpy.ndarray`
        The Gaussian function evaluated at provided x.
    """

    return a * np.exp(-(x - x0)**2 / (2 * sigma**2))


def CircularGaussian(xy, a, x0, y0, sigma, b):
    """Circular Gaussian function.

    Parameters
    ----------
    xy : scalar or `~numpy.ndarray`
        Value or values to evaluate the Gaussian at.

    a : scalar
        Amplitude.

    x0, y0 : scalar
        Gaussian center.

    sigma : scalar
        Standard deviation.

    b : scalar
        Offset.

    Returns
    -------
    out: scalar or `~numpy.ndarray`
        The circular Gaussian function evaluated at provided (x, y).
    """

    x, y = xy

    return a * np.exp(-(x - x0)**2 / (2 * sigma**2) - (y - y0)**2 / (2 * sigma**2)) + b


import warnings

# AstroPy
from astropy.time import Time
# ccdproc
from ccdproc.utils.slices import slice_from_string
# drpy
from drpy.validate import _validate1DArray, _validateCCDList

__all__ = ['concatenate']


def concatenate(ccdlist, fits_section, scale=None):
    """Concatenate two frames.

    The region defined by ``fits_section`` of the second frame is replaced by that in 
    the first frame. Usually this is used to concatenate flat-fields or arc spectra of 
    different exposure times. To concatenate multiple frames, call this function 
    repeatedly.

    Parameters
    ----------
    ccdlist : Iterable
        Iterable object containing two 2-dimensional frames.

    fits_section : str
        Region of the second frame to be replaced by the counterpart of the first frame.
    
    scale : scalar, array-like or `None`, optional
        Scaling factor. Frames are multiplied by scaling factor prior to concatenation.
        Default is `None`.

    Returns
    -------
    nccd : `~astropy.nddata.CCDData`
        The concatenated frame.
    """

    ccdlist = _validateCCDList(ccdlist, 2, 2)
    n_image = len(ccdlist)
    if n_image > 2:
        warnings.warn(
            '``ccdlist`` contains more than two frames. Extra frames will be ignored.', 
            RuntimeWarning)

    python_slice = slice_from_string(fits_section, fits_convention=True)

    if scale is None:
        scale = _validate1DArray(1, 'scale', 2, True)

    else:
        scale = _validate1DArray(scale, 'scale', 2, True)

    data_arr = ccdlist[1].data * scale[1]
    data_arr[python_slice] = ccdlist[0].data[python_slice] * scale[0]

    if (ccdlist[0].uncertainty is None) | (ccdlist[1].uncertainty is None):
        uncertainty_arr = None

    else:
        uncertainty_arr = scale[1] * ccdlist[1].uncertainty.array
        uncertainty_arr[python_slice] = (
            ccdlist[0].uncertainty.array[python_slice] * scale[0])

    if (ccdlist[0].mask is None) | (ccdlist[1].mask is None):
        mask_arr = None

    else:
        mask_arr = ccdlist[1].mask.copy()
        mask_arr[python_slice] = ccdlist[0].mask[python_slice]

    nccd = ccdlist[0].copy()

    nccd.data = data_arr

    if uncertainty_arr is None:
        nccd.uncertainty = None

    else:
        # Here ``nccd.uncertainty`` is definitely not `None`.
        nccd.uncertainty.array = uncertainty_arr

    nccd.mask = mask_arr

    nccd.header['CONCATEN'] = (
        '{} Data section {} is from the first frame.'.format(
            Time.now().to_value('iso', subfmt='date_hm'), fits_section)
    )

    return nccd


"""
drpy: a data reduction toolkit for astronomical photometry and spectroscopy
"""

import os

# AstroPy
from astropy import config as _config

__all__ = ['__version__', 'conf']

__version__ = '0.0.1.8'


class Conf(_config.ConfigNamespace):
    """Configuration parameters for `drpy`."""

    # Unit
    unit_ccddata = _config.ConfigItem(
        'adu', 'Default unit for `~astropy.nddata.CCDData` objects.')

    unit_spectral_axis = _config.ConfigItem(
        'pixel', 'Default unit for `~specutils.Spectrum1D` objects.')

    unit_flux = _config.ConfigItem(
        'count', 'Default unit for `~specutils.Spectrum1D` objects.')
    
    # Data type
    dtype_ccddata = _config.ConfigItem(
        'float32', 'Default dtype for `~astropy.nddata.CCDData` objects.')

    # Plot
    fig_width = _config.ConfigItem(6, 'Width of figure.')
    
    fig_show = _config.ConfigItem(False, 'Whether to show plots.')

    fig_save = _config.ConfigItem(False, 'Whether to save plots.')

    fig_path = _config.ConfigItem(os.getcwd(), 'Where to save plots.')


conf = Conf()

del _config


class Tracer:
    
    def __init__(self, ):
        
        pass
    
    def _trace():
        
        pass
    
    def trace():
        
        pass
    
    def plot():
        
        pass
    
    def writeto():
        
        pass


import warnings


def ignoreWarning(func, action, category, lineno=0, append=False):
    
    def wrapper(*args, **kwargs):
        
        with warnings.catch_warnings():
            
            # Filter warnings
            warnings.simplefilter(action, category, lineno, append)
            
            return func(*args, **kwargs)
    
    return wrapper


from copy import deepcopy
import os

# NumPy
import numpy as np
# matplotlib
from matplotlib import gridspec
import matplotlib.pyplot as plt
# AstroPy
from astropy.visualization import ZScaleInterval
# drpy
from drpy import conf

from .validate import (_validateBool, _validatePath, _validateRange, _validateString, 
                       _validateSpectrum)

# Set plot parameters
plt.rcParams['figure.figsize'] = [conf.fig_width, conf.fig_width]
plt.rcParams['axes.linewidth'] = 1.5
plt.rcParams['mathtext.fontset'] = 'stix'
plt.rcParams['font.family'] = 'STIXGeneral'

__all__ = ['plotFitting', 'plotSpectrum1D', 'plot2d']


def _plotFitting(ax, x, y, m, x_fit, y_fit, r, threshold_lower, threshold_upper, 
                 xlabel='x', ylabel='y', use_relative=False):
    """Plot fitting."""

    # ymin_fit = np.nanmin(y_fit)
    # ymax_fit = np.nanmax(y_fit)
    # y_fit_range = ymax_fit - ymin_fit
    # ymin_fit -= 0.1 * y_fit_range
    # ymax_fit += 0.1 * y_fit_range

    if (threshold_lower is None) & (threshold_upper is None):
        ymin = None; ymax = None
    elif threshold_lower is None:
        ymax = 1.67 * threshold_upper; ymin = -ymax
    elif threshold_upper is None:
        ymin = 1.67 * threshold_lower; ymax = -ymin
    else:
        ymax = 1.67 * np.max(np.abs([threshold_lower, threshold_upper])); ymin = -ymax

    rms = np.sqrt((r[~m]**2).sum() / (~m).sum())

    # Data
    ax[0].plot(x[~m], y[~m], '+', c='black', ms=8, zorder=-2)
    ax[0].plot(x[m], y[m], '+', c='lightgrey', ms=8, zorder=-3)

    # Settings
    ax[0].set_xlim(x_fit[0], x_fit[-1])
    ax[0].autoscale(False)

    # Fitted data
    ax[0].plot(x_fit, y_fit, '-', c='red', lw=1.5, zorder=-1)
    
    # Settings
    ax[0].grid(True, ls='--')
    # ticks
    ax[0].tick_params(
        which='major', direction='in', top=True, right=True, length=5, width=1.5, 
        labelsize=12)
    ax[0].minorticks_off()
    # labels
    ax[0].set_ylabel(ylabel, fontsize=16)

    # Residual
    ax[1].plot(x[m], r[m], '+', c='lightgrey', ms=8)
    ax[1].plot(x[~m], r[~m], '+', c='black', ms=8)
    ax[1].axhline(y=0, ls='-', c='red', lw=1.5)
    if threshold_lower is not None:
        ax[1].axhline(y=threshold_lower, ls='--', c='red', lw=1.5)
    if threshold_upper is not None:
        ax[1].axhline(y=threshold_upper, ls='--', c='red', lw=1.5)
    
    # Settings
    ax[1].grid(True, ls='--')
    # ticks
    ax[1].tick_params(
        which='major', direction='in', top=True, right=True, length=5, width=1.5, 
        labelsize=12)
    ax[1].minorticks_off()
    # lim
    ax[1].set_ylim(ymin, ymax)
    # label
    ax[1].set_xlabel(xlabel, fontsize=16)
    if use_relative:
        label_residual = 'rel. residual'
    else:
        label_residual = 'residual'
    ax[1].set_ylabel(label_residual, fontsize=16)
    ax[1].annotate(
        f'RMS$={rms:.3}$', (0.97, 0.8), xycoords='axes fraction', ha='right', 
        fontsize=12)


def plotFitting(x, y, m, y_fit, r, x_fit=None, threshold_lower=None, 
                threshold_upper=None, xlabel='x', ylabel='y', title='fitting', 
                show=conf.fig_show, save=conf.fig_save, path=conf.fig_path, 
                use_relative=False):
    """Plot fitting.

    Parameters
    ----------
    show : bool, optional
        Whether to show plot.

    save : bool, optional
        Whether to save plot.

    path : bool, optional
        Path to save plot in.
    """

    _validateBool(show, 'show')
    _validateBool(save, 'save')

    if show | save:

        _validateString(title, 'title')
        if title != 'fitting':
            title = f'{title} fitting'

        if x_fit is None:
            x_fit = deepcopy(x)

        if threshold_lower is not None:
            if np.isnan(threshold_lower) | np.isinf(threshold_lower):
                threshold_lower = None
            else:
                _validateRange(
                    threshold_lower, 'threshold_lower', (None, 0), (None, True))

        if threshold_upper is not None:
            if np.isnan(threshold_upper) | np.isinf(threshold_upper):
                threshold_upper = None
            else:
                _validateRange(
                    threshold_upper, 'threshold_upper', (0, None), (True, None))

        _validateBool(use_relative, 'use_relative')

        fig, ax = plt.subplots(2, 1, sharex=True, height_ratios=[3, 1], dpi=100)
        fig.subplots_adjust(hspace=0)

        _plotFitting(
            ax, x, y, m, x_fit, y_fit, r, threshold_lower, threshold_upper, xlabel, 
            ylabel, use_relative)

        ax[0].set_title(title, fontsize=16)
        fig.align_ylabels()
        fig.tight_layout()
        
        if save:
            fig_path = _validatePath(path, title)
            plt.savefig(fig_path, dpi=100)

        if show:
            plt.show()

        plt.close()


def _plotSpectrum1D(ax, spectral_axis, flux, uncertainty=None, xlabel='spectral axis', 
                    ylabel='flux'):
    """Plot 1-dimensional spectrum of type `~specutils.Spectrum1D`."""

    # 1-dimension
    if flux.ndim == 1:
        ax.step(spectral_axis, flux, where='mid', color='C0', lw=1.5, zorder=2.5)
        (ymin, ymax) = ax.get_ylim()
        if uncertainty is not None:
            ax.fill_between(
                spectral_axis, flux - uncertainty, flux + uncertainty, color='C0', 
                alpha=0.2, zorder=2.5)
    # more than 1-dimension
    else:
        flux = flux.reshape(-1, flux.shape[-1])
        for i in range(flux.shape[0]):
            ax.step(spectral_axis, flux[i], where='mid', color='C0', lw=1.5, zorder=2.5)
        (ymin, ymax) = ax.get_ylim()
        for i in range(flux.shape[0]):
            if uncertainty[i] is not None:
                ax.fill_between(
                    spectral_axis, (flux - uncertainty)[i], (flux + uncertainty)[i], 
                    color='C0', alpha=0.2, zorder=2.5)

    # Settings
    ax.grid(axis='both', color='0.95', zorder=-1)
    ax.set_xlim(spectral_axis[0], spectral_axis[-1])
    ax.set_ylim(ymin, ymax)
    ax.tick_params(
        which='major', direction='in', top=True, right=True, length=5, width=1.5, 
        labelsize=12)
    ax.minorticks_off()
    ax.set_xlabel(xlabel, fontsize=16)
    ax.set_ylabel(ylabel, fontsize=16)
    
# todo: label masked pixels, add legend?
def plotSpectrum1D(spectrum1d, title='spectrum', show=conf.fig_show, 
                   save=conf.fig_save, path=conf.fig_path):
    """Plot 1-dimensional spectrum of type `~specutils.Spectrum1D`.
    
    Parameters
    ----------
    spectrum1d : `~specutils.Spectrum1D` or `~numpy.ndarray`
        Input spectrum.

    show : bool, optional
        Whether to show plot.

    save : bool, optional
        Whether to save plot.

    path : bool, optional
        Path to save plot in.
    """

    _validateBool(show, 'show')
    _validateBool(save, 'save')

    if show | save:

        _validateString(title, 'title')
        if title != 'spectrum':
            title = f'{title} spectrum'

        new_spectrum1d, spectral_axis, flux, uncertainty, mask = _validateSpectrum(
            spectrum1d, 'spectrum1d', True, True)

        # Labels
        unit_spectral_axis = new_spectrum1d.spectral_axis.unit.to_string()
        unit_flux = new_spectrum1d.flux.unit.to_string()

        xlabel = f'spectral axis [{unit_spectral_axis}]'
        ylabel = f'flux [{unit_flux}]'

        fig, ax = plt.subplots(1, 1, dpi=100)
        _plotSpectrum1D(
            ax, spectral_axis, flux, uncertainty, xlabel=xlabel, ylabel=ylabel)
        # ax.legend(fontsize=16)
        ax.set_title(title, fontsize=16)
        fig.set_figheight(0.7 * fig.get_figwidth())
        fig.tight_layout()

        if save:
            fig_path = _validatePath(path, title)
            plt.savefig(fig_path, dpi=100)

        if show:
            plt.show()

        plt.close()


def _plot2d(ax, ccd, cmap='Greys_r', contrast=0.25, cbar=True, xlabel='column', 
            ylabel='row', cblabel='pixel value', **kwargs):
    """Plot image."""

    if ('vmin' not in kwargs) | ('vmax' not in kwargs):
        # The first argument is ``n_samples`` after version 5.2 of AstroPy while 
        # ``nsamples`` before that. Here it is used as positional argument out of 
        # compatibility.
        zscale = ZScaleInterval(
            600, contrast=contrast, max_reject=0.5, min_npixels=5, krej=2.5, 
            max_iterations=5)
        if 'vmax' in kwargs:
            kwargs['vmin'], _ = zscale.get_limits(ccd)
        elif 'vmin' in kwargs:
            _, kwargs['vmax'] = zscale.get_limits(ccd)
        else:
            kwargs['vmin'], kwargs['vmax'] = zscale.get_limits(ccd)

    # Image
    im = ax.imshow(ccd, cmap=cmap, origin='lower', **kwargs)
    # Settings
    ax.tick_params(
        which='major', direction='in', top=True, right=True, color='w', length=5, 
        width=1.5, labelsize=12)
    ax.minorticks_off()
    ax.set_xlabel(xlabel, fontsize=16)
    ax.set_ylabel(ylabel, fontsize=16)

    # Colorbar
    if cbar:
        cb = plt.colorbar(
            im, ax=ax, fraction=(0.046 * ccd.shape[0] / ccd.shape[1]), pad=0.04, 
            use_gridspec=True)
        # Settings
        cb.ax.tick_params(
            which='major', direction='in', right=True, color='w', length=5, width=1.5, 
            labelsize=12)
        cb.ax.set_ylabel(cblabel, fontsize=16)


def plot2d(ccd, cmap='Greys_r', contrast=0.25, cbar=True, xlabel='column', 
           ylabel='row', cblabel='pixel value', title='image', show=conf.fig_show, 
           save=conf.fig_save, path=conf.fig_path, **kwargs):
    """Plot image.

    Parameters
    ----------
    show : bool, optional
        Whether to show plot.

    save : bool, optional
        Whether to save plot.

    path : bool, optional
        Path to save plot in.
    """

    _validateBool(show, 'show')
    _validateBool(save, 'save')

    if show | save:

        _validateString(title, 'title')

        fig, ax = plt.subplots(1, 1, dpi=100)
        _plot2d(
            ax=ax, ccd=ccd, cmap=cmap, contrast=contrast, cbar=cbar, xlabel=xlabel, 
            ylabel=ylabel, cblabel=cblabel, **kwargs)
        ax.set_title(title, fontsize=16)
        fig.set_figheight(ccd.shape[0] / ccd.shape[1] * fig.get_figwidth())
        fig.tight_layout()

        if save:
            fig_path = _validatePath(path, title)
            plt.savefig(fig_path, dpi=100)

        if show:
            plt.show()

        plt.close()